# encoding: utf-8

#* Copyright (C) 2019 CZ.NIC
#*
#* This program is free software: you can redistribute it and/or modify
#* it under the terms of the GNU General Public License as published by
#* the Free Software Foundation, either version 3 of the License, or
#* (at your option) any later version.
#*
#* This program is distributed in the hope that it will be useful,
#* but WITHOUT ANY WARRANTY; without even the implied warranty of
#* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#* GNU General Public License for more details.
#*
#* You should have received a copy of the GNU General Public License
#* along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Python high level API for libisds.

It requires at least Python 3.7.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum, unique
from datetime import date
from typing import Iterable

import isds


##########################################################################################
# Exceptions

@unique
class ErrorCode(IntEnum):
    """Enumeration with ISDS error codes."""

    SUCCESS = isds.IE_SUCCESS
    ERROR = isds.IE_ERROR
    NOTSUP = isds.IE_NOTSUP
    INVAL = isds.IE_INVAL
    INVALID_CONTEXT = isds.IE_INVALID_CONTEXT
    NOT_LOGGED_IN = isds.IE_NOT_LOGGED_IN
    CONNECTION_CLOSED = isds.IE_CONNECTION_CLOSED
    TIMED_OUT = isds.IE_TIMED_OUT
    NOEXIST = isds.IE_NOEXIST
    NOMEM = isds.IE_NOMEM
    NETWORK = isds.IE_NETWORK
    HTTP = isds.IE_HTTP
    SOAP = isds.IE_SOAP
    XML = isds.IE_XML
    ISDS = isds.IE_ISDS
    ENUM = isds.IE_ENUM
    DATE = isds.IE_DATE
    TOO_BIG = isds.IE_2BIG
    TOO_SMALL = isds.IE_2SMALL
    NOTUNIQ = isds.IE_NOTUNIQ
    NOTEQUAL = isds.IE_NOTEQUAL
    PARTIAL_SUCCESS = isds.IE_PARTIAL_SUCCESS
    ABORTED = isds.IE_ABORTED
    SECURITY = isds.IE_SECURITY


class IsdsError(Exception):
    """
    Generic error from libisds.

    Attributes:
        code: Exception error code.
        code_descr: Human readable error code meaning.
        message: Human readable string describing the exception.
    """

    def __init__(self, code: ErrorCode, code_descr: str, message: str):
        self.code = code
        self.code_descr = code_descr
        self.message = message
        super().__init__(code, code_descr, message)


class ConnectionFailed(IsdsError):
    """Connection to ISDS failed."""


class AuthenticationFailed(IsdsError):
    """Authentication to ISDS failed."""

class TotpCodeRequired(IsdsError):
    """TOTP code has been sent through a side channel and is now required to complete the login procedure."""


##########################################################################################
# Helper classes
@dataclass
class Credentials(ABC):
    """
    Base class for credential classes.

    Attributes:
        username: User login name.
    """

    username: str

    @abstractmethod
    def login(self, connection: isds.isds_ctx, *, test: bool = False) -> None:
        """
        Login to connection.

        Derived class must implement this method.

        Raises:
            AuthenticationFailed: If authentication fails.
        """


@dataclass
class Password(Credentials):
    """
    Login with plain password.

    Attributes:
        password: User login password.
    """

    password: str

    def login(self, connection: isds.isds_ctx, *, test: bool = False) -> None:
        """
        Login to connection with password.

        Raises:
            AuthenticationFailed: If authentication fails.
        """
        result = connection.login_pwd(self.username, password=self.password, test_env=test)
        if result != ErrorCode.SUCCESS:
            raise AuthenticationFailed(result, isds.isds_strerror(result), connection.long_message())


@unique
class KeyFormat(IntEnum):
    """Enumeration with certificate key formats."""

    PEM = isds.PKI_FORMAT_PEM
    DER = isds.PKI_FORMAT_DER
    ENG = isds.PKI_FORMAT_ENG


@dataclass
class Certificate(Credentials):
    """
    Login with certificate.

    Attributes:
        password: User password. You don't have to supply the password if the account just uses the certificate to log in.
        cert_format: Certificate format.
        cert: Path to client certificate.
        key_format: Private key format.
        key: Path to client private key.
        passphrase: String with password for private key decryption. Use 'None' for no pass-phrase.
    """

    password: str = None
    cert_format: KeyFormat = KeyFormat.PEM
    cert: str  = None
    key_format: KeyFormat = KeyFormat.PEM
    key: str = None
    passphrase: str = None

    def login(self, connection: isds.isds_ctx, *, test: bool = False) -> None:
        """
        Login to connection with certificate.

        Raises:
            AuthenticationFailed: If authentication fails.
        """
        result = connection.login_cert(self.username, self.password,
            cert_format=self.cert_format, cert=self.cert, key_format=self.key_format, key=self.key,
            passphrase=self.passphrase, test_env=test)
        if result != ErrorCode.SUCCESS:
            raise AuthenticationFailed(result, isds.isds_strerror(result), connection.long_message())


@dataclass
class Hotp(Credentials):
    """
    HMAC-based OTP login.

    Attributes:
        password: User password.
        hotp_code: HMAC-based OTP code.
    """

    password: str
    hotp_code: str

    def login(self, connection: isds.isds_ctx, *, test: bool = False) -> None:
        """
        Login to connection using a HMAC-based OTP code.

        Raises:
            AuthenticationFailed: If authentication fails.
        """
        result,otp_result = connection.login_hotp(self.username, self.password, self.hotp_code, test_env=test)
        if result != ErrorCode.SUCCESS:
            raise AuthenticationFailed(result, isds.isds_strerror(result), connection.long_message())


@dataclass
class Totp(Credentials):
    """
    Time-based OTP login.

    Attributes:
        password: User password.
        totp_code: Time-based OTP code (SMS code).
    """

    password: str
    totp_code: str = None

    def login(self, connection: isds.isds_ctx, *, test: bool = False) -> None:
        """
        Login to connection using a HMAC-based OTP code.

        Raises:
            TotpCodeRequired: If the request for the SMS code succeeded. The application is expected to fill in the obtained code into `totp_code` and retry the login.
            AuthenticationFailed: If authentication fails.
        """
        result,otp_result = connection.login_totp(self.username, self.password, self.totp_code, test_env=test)
        if result == ErrorCode.PARTIAL_SUCCESS:
            raise TotpCodeRequired(result, isds.isds_strerror(result), connection.long_message())
        elif result != ErrorCode.SUCCESS:
            raise AuthenticationFailed(result, isds.isds_strerror(result), connection.long_message())

@unique
class DataBoxType(IntEnum):
    """Enumeration with data box types."""

    OVM_MAIN = isds.DBTYPE_OVM_MAIN
    SYSTEM = isds.DBTYPE_SYSTEM
    OVM = isds.DBTYPE_OVM
    OVM_NOTAR = isds.DBTYPE_OVM_NOTAR
    OVM_EXEKUT = isds.DBTYPE_OVM_EXEKUT
    OVM_REQ = isds.DBTYPE_OVM_REQ
    OVM_FO = isds.DBTYPE_OVM_FO
    OVM_PFO = isds.DBTYPE_OVM_PFO
    OVM_PO = isds.DBTYPE_OVM_PO
    PO = isds.DBTYPE_PO
    PO_ZAK = isds.DBTYPE_PO_ZAK
    PO_REQ = isds.DBTYPE_PO_REQ
    PFO = isds.DBTYPE_PFO
    PFO_ADVOK = isds.DBTYPE_PFO_ADVOK
    PFO_DANPOR = isds.DBTYPE_PFO_DANPOR
    PFO_INSSPR = isds.DBTYPE_PFO_INSSPR
    PFO_AUDITOR = isds.DBTYPE_PFO_AUDITOR
    FO = isds.DBTYPE_FO


@dataclass
class _PersonName:
    """Name of a person."""

    first_name: str = None
    middle_name: str = None
    last_name: str = None
    last_name_at_birth: str = None


@dataclass
class _BirthInfo:
    """Information about birth."""

    birth_date: date = None
    birth_city: str = None
    birth_country: str = None
    birth_state: str = None


@dataclass
class _Address:
    """Information about address."""

    street: str = None
    conscription_number: str = None  # adNumberInMunicipality
    orientation_number: str = None  # adNumberInStreet
    city: str = None
    zip_code: str = None
    state: str = None


@dataclass
class DataBoxOwner(_PersonName, _BirthInfo, _Address):
    """Data box owner info."""

    ic: str = None
    firm_name: str = None
    nationality: str = None
    email: str = None
    telephone: str = None
    identifier: str = None
    registry_code: str = None
    db_id: str = None
    db_type: DataBoxType = None
    db_state: int = None
    is_effective_ovm: bool = None  # dbEffectiveOVM
    has_open_addressing: bool = None  # dbOpenAddressing


class DataBoxList(list):
    """
    List of data box owners.

    Attributes:
        truncated: Whether the list was truncated.
    """

    def __init__(self, iterable: Iterable[DataBoxOwner] = None, *, truncated: bool = False):
        super().__init__(iterable or ())
        self.truncated = truncated


##########################################################################################
# Client

# TODO: Ensure thread safety.
# TODO: If the connection fails, try to reconnect.  - Reconnection using HOTP
# and TOTP requires user interaction. TOTP authentication uses premium SMS
# messages which cost money. Reconnection using a certificate may in some cases
# also require user interaction to decrypt the private key.
class Client:
    """
    ISDS client.

    Connects and authenticates lazily, i.e. when needed.

    Attributes:
        credentials: User credentials.
        timeout: Timeout in seconds.
        test: If True, connects to testing environment.
    """

    def __init__(self, credentials: Credentials, *, timeout: int = None, test: bool = False):
        self.credentials = credentials
        self.timeout = timeout
        self.test = test
        self._connection = None

    def __del__(self):
        """
        Log out from ISDS.

        Ignores any errors from log out.
        """
        if self._connection is not None:
            # Ignore all errors from logout.
            self._connection.logout()

    @property
    def connection(self):
        """Return connection object.

        If not connected, creates a connection.
        """
        if self._connection is None:
            self.connect()
        return self._connection

    def connect(self) -> None:
        """Connect to the ISDS.

        Raises:
            ConnectionFailed: If connection failed.
            AuthenticationFailed: If authentication failed.
        """
        self._connection = isds.isds_ctx()
        if self.timeout:
            result = self._connection.set_timeout(self.timeout * 1000)
            if result != ErrorCode.SUCCESS:
                raise ConnectionFailed(result, isds.isds_strerror(result), self._connection.long_message())
        self.credentials.login(self._connection, test=self.test)

    def ping(self) -> None:
        """
        Check if connection is still alive.

        Raises:
            ConnectionFailed: If connection is no longer active.
        """
        result = self.connection.ping()
        if result != ErrorCode.SUCCESS:
            raise ConnectionFailed(result, isds.isds_strerror(result), self._connection.long_message())

    def find_box(
            self, db_type: DataBoxType, *, first_name: str = None, middle_name: str = None, last_name: str = None,
            last_name_at_birth: str = None, birth_date: date = None, birth_city: str = None, birth_country: str = None,
            birth_state: str = None, street: str = None, conscription_number: str = None,
            orientation_number: str = None, city: str = None, zip_code: str = None, state: str = None,
            db_id: str = None, ic: str = None, firm_name: str = None, nationality: str = None, email: str = None,
            telephone: str = None, identifier: str = None, registry_code: str = None, db_state: int = None,
            is_effective_ovm: bool = None, has_open_addressing: bool = None) -> DataBoxList:
        """Find data boxes based on filter criteria."""
        query = isds.isds_db_owner_info()
        if first_name is not None or middle_name is not None or last_name is not None or last_name_at_birth is not None:
            query.personName = isds.isds_person_name(pnFirstName=first_name, pnMiddleName=middle_name,
                                                     pnLastName=last_name, pnLastNameAtBirth=last_name_at_birth)
        if birth_date is not None or birth_city is not None or birth_country is not None or birth_state is not None:
            birth_info = isds.isds_birth_info(biCity=birth_city, biCounty=birth_country, biState=birth_state)
            if birth_date is not None:
                birth_info.biDate = isds.isds_date(birth_date.year, birth_date.month, birth_date.day)
            query.birthInfo = birth_info
        if (street is not None or conscription_number is not None or orientation_number is not None or city is not None
                or zip_code is not None or state is not None):
            query.address = isds.isds_address(
                adStreet=street, adNumberInMunicipality=conscription_number, adNumberInStreet=orientation_number,
                adCity=city, adZipCode=zip_code, adState=state)

        query.dbID = db_id
        query.dbType = db_type
        query.ic = ic
        query.firmName = firm_name
        query.nationality = nationality
        query.email = email
        query.telNumber = telephone
        query.identifier = identifier
        query.registryCode = registry_code
        query.dbState = db_state
        query.dbEffectiveOVM = is_effective_ovm
        query.dbOpenAddressing = has_open_addressing

        return_code, result = self.connection.find_box(query)
        if return_code == ErrorCode.NOEXIST:
            # No data box was found, return empty list.
            return DataBoxList()
        if return_code not in (ErrorCode.SUCCESS, ErrorCode.TOO_BIG):
            raise IsdsError(return_code, self.connection.long_message())
        data_boxes = DataBoxList(truncated=(return_code == ErrorCode.TOO_BIG))
        for item in result:
            data_box = DataBoxOwner(
                db_id=item.dbID,
                db_type=item.dbType,
                ic=item.ic,
                firm_name=item.firmName,
                nationality=item.nationality,
                email=item.email,
                telephone=item.telNumber,
                identifier=item.identifier,
                registry_code=item.registryCode,
                db_state=item.dbState,
                is_effective_ovm=item.dbEffectiveOVM,
                has_open_addressing=item.dbOpenAddressing,
            )
            person_name = item.personName
            if person_name is not None:
                data_box.first_name = person_name.pnFirstName
                data_box.middle_name = person_name.pnMiddleName
                data_box.last_name = person_name.pnLastName
                data_box.last_name_at_birth = person_name.pnLastNameAtBirth
            del person_name
            birth_info = item.birthInfo
            if birth_info is not None:
                data_box.birth_city = birth_info.biCity
                data_box.birth_country = birth_info.biCounty
                data_box.birth_state = birth_info.biState
                bi_date = birth_info.biDate
                if birth_info.biDate:
                    data_box.birth_date = date(bi_date.year, bi_date.mon, bi_date.mday)
                del bi_date
            del birth_info
            address = item.address
            if address is not None:
                data_box.street = address.adStreet
                data_box.conscription_number = address.adNumberInMunicipality
                data_box.orientation_number = address.adNumberInStreet
                data_box.city = address.adCity
                data_box.zip_code = address.adZipCode
                data_box.state = address.adState
            del address
            data_boxes.append(data_box)
        return data_boxes
